// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import java.io.*;
import java.math.BigInteger;

/**
 * Logs the runtime activity of a thread into a file.
 *
 * @author Esko Luontola
 */
public class ProfilerLogger {

    public static final File LOG_DIRECTORY = new File("profiler_logs");
    public static final String LOG_PREFIX = "thread-";
    public static final String LOG_SUFFIX = ".log";

    static {
        LOG_DIRECTORY.mkdirs();
        if (!LOG_DIRECTORY.isDirectory()) {
            throw new IllegalStateException("Logging directory does not exists: " + LOG_DIRECTORY);
        }

        // TODO: Install a default exception handler to avoid incomplete log files
        // (will no fix the same problem with System.exit)
        // http://www-128.ibm.com/developerworks/java/library/j-tiger08104/?ca=dgr-UVs16-JDK_5.0-Threads
    }

    private static BigInteger time = BigInteger.ZERO;
    private static long lastNanoTime = System.nanoTime();

    private static ThreadLocal<ThreadState> threadLocal = new ThreadLocal<ThreadState>() {
        @Override
        protected ThreadState initialValue() {
            return new ThreadState();
        }
    };

    private static class ThreadState {

        private int reclevel;

        private FileWriter writer;

        public ThreadState() {
            reclevel = -1;
            File file;
            int i = 0;
            do {
                file = new File(LOG_DIRECTORY, LOG_PREFIX + i + LOG_SUFFIX);
                i++;
            } while (file.exists());
            try {
                writer = new FileWriter(file);
            } catch (IOException e) {
                throw new IllegalStateException("Unable to write file: " + file, e);
            }
        }

        public int getReclevel() {
            return reclevel;
        }

        public void setReclevel(int reclevel) {
            this.reclevel = reclevel;
        }

        public FileWriter getWriter() {
            return writer;
        }
    }

    public static int logEnter(String classID, String methodID) {
        enterProfiler();
        int reclevel = getNewRecLevel();
        writeEntry(getTime() + ";ENTER;" + reclevel + ";" + classID + ";" + methodID);
        exitProfiler();
        return reclevel;
    }

    public static void logCatch(int reclevel, String classID, String methodID) {
        enterProfiler();
        setRecLevel(reclevel);
        writeEntry(getTime() + ";CATCH;" + reclevel + ";" + classID + ";" + methodID);
        exitProfiler();
    }

    public static void logReturn(int reclevel, String classID, String methodID) {
        enterProfiler();
        setRecLevel(reclevel - 1);
        writeEntry(getTime() + ";RETURN;" + reclevel + ";" + classID + ";" + methodID);
        exitProfiler();
    }

    private static void writeEntry(String entry) {
        ThreadState state = threadLocal.get();
        try {
            state.getWriter().append(entry + "\n");
            state.getWriter().flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private static void enterProfiler() {
        long nanoTime = System.nanoTime();
        time = time.add(BigInteger.valueOf(nanoTime - lastNanoTime));
    }

    private static void exitProfiler() {
        lastNanoTime = System.nanoTime();
    }

    private static BigInteger getTime() {
        return time;
    }

    public static int getNewRecLevel() {
        ThreadState state = threadLocal.get();
        int reclevel = state.getReclevel() + 1;
        state.setReclevel(reclevel);
        return reclevel;
    }

    private static void setRecLevel(int reclevel) {
        ThreadState state = threadLocal.get();
        state.setReclevel(reclevel);
    }
}
