// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import net.orfjackal.util.Tools;
import org.jetbrains.annotations.NotNull;

import java.util.*;

/**
 * Represents a string replace command that will be used in connection with other replace commands.
 *
 * @author Esko Luontola
 */
public class StringReplacement implements Comparable<StringReplacement> {

    private int start;

    private int end;

    @NotNull
    private String replacement;

    public StringReplacement(int pos, String replacement) {
        this(pos, pos, replacement);
    }

    public StringReplacement(int start, int end, String replacement) {
//        System.out.println(start + " " + end + " " + replacement);
        if (start < 0) {
            throw new IllegalArgumentException("start = " + start + " < 0");
        } else if (end < 0) {
            throw new IllegalArgumentException("end = " + end + " < 0");
        } else if (start > end) {
            throw new IllegalArgumentException("start = " + start + " > end = " + end);
        } else if (replacement == null) {
            throw new NullPointerException("replacement is null");
        }
        this.start = start;
        this.end = end;
        this.replacement = replacement;
    }

    public int getStart() {
        return start;
    }

    public int getEnd() {
        return end;
    }

    @NotNull
    public String getReplacement() {
        return replacement;
    }

    public int compareTo(StringReplacement o) {
        return this.start - o.start;
    }

    public static String replace(String source, List<StringReplacement> replacements) {
        Collections.sort(replacements);
        StringBuilder sb = new StringBuilder(source.length());

        // remove duplicates from the replacements
        for (int i = 0; i < replacements.size() - 1; i++) {
            if (replacements.get(i).getEnd() > replacements.get(i + 1).getStart()) {
                StringReplacement r = replacements.get(i + 1);
                if (Tools.isDebug()) {
                    System.err.println("Removed replacement: start " + r.getStart() + ", end " + r.getEnd()
                            + ", " + r.getReplacement());
                }
                replacements.remove(i + 1);
            }
        }

        int pos = 0;
        for (StringReplacement rep : replacements) {
            sb.append(source, pos, rep.getStart());
            sb.append(rep.getReplacement());
            pos = rep.getEnd();
        }
        sb.append(source, pos, source.length());

        return sb.toString();
    }
}
