// Copyright © 2005, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.util.commandline;

import org.jetbrains.annotations.NotNull;

import java.text.ParseException;

/**
 * Represents one command line argument that a command line application can take. Arguments are always written after the
 * option parameters on command line.
 *
 * @author Esko Luontola
 */
public abstract class Argument {

    @NotNull
    private String name;

    private boolean required;

    private boolean multiple;

    /**
     * Creates a new command line argument that can be given only once.
     *
     * @param name     the name of the argument, for documentation purposes. May contain uppercase alphabetical
     *                 characters and underscores.
     * @param required true if the argument is mandatory, otherwise false.
     * @throws NullPointerException if name is null.
     */
    public Argument(String name, boolean required) {
        this(name, required, false);
    }

    /**
     * Creates a new command line argument.
     *
     * @param name     the name of the argument, for documentation purposes. May contain uppercase alphabetical
     *                 characters and underscores.
     * @param required true if the argument is mandatory, otherwise false.
     * @param multiple true if many instances of this argument can be given.
     * @throws NullPointerException if name is null.
     */
    public Argument(String name, boolean required, boolean multiple) {
        if (name == null) {
            throw new NullPointerException("name is null");
        } else if (!name.matches("[A-Z_]+")) {
            throw new IllegalArgumentException(
                    "name must contain one or more uppercase alphabetical characters (A-Z) or underscore (_). It was: " + name);
        }
        this.name = name;
        this.required = required;
        this.multiple = multiple;
    }

    @NotNull
    public String getName() {
        return name;
    }

    public boolean isRequired() {
        return required;
    }

    public boolean isMultiple() {
        return multiple;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder(name.length() + 5);
        if (!isRequired()) {
            sb.append('[');
        }
        sb.append(getName());
        if (!isRequired()) {
            sb.append(']');
        }
        if (isMultiple()) {
            sb.append("...");
        }
        return sb.toString();
    }

    /**
     * This method will be called by the CommandLineParser once, if this argument has been specified on the command
     * line. The subclasses should implement this method to do the actions that will reflect the argument.
     *
     * @param args the arguments that were given on command line. Each argument is its own string and length of the
     *             array must be at least one. The method is allowed to change the contents of the array.
     * @throws NullPointerException           if args is null.
     * @throws ArrayIndexOutOfBoundsException if the length of args is zero.
     * @throws ParseException                 if there is an unknown parameter. The message of the exception will
     *                                        contain the reason.
     */
    public abstract void onSpecified(String[] args) throws ParseException;

}
