// Copyright © 2005-2009, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import org.junit.*;
import org.mockito.Mockito;

import java.io.*;

/**
 * @author Esko Luontola
 */
public class FileAnalyzerTest extends Assert {

    private final AnalyzerBatch mockAnalyzerBatch = Mockito.mock(AnalyzerBatch.class);
    private final File dummyFile = new File("dummy.file");

    @After
    public void no_more_AnalyzerBatch_interactions() {
        Mockito.verifyNoMoreInteractions(mockAnalyzerBatch);
    }

    private FileAnalyzer createFileAnalyzer(final String fileContents) throws IOException {
        return new FileAnalyzer(mockAnalyzerBatch, dummyFile) {
            protected String readFile(File file) {
                assertEquals(dummyFile, file);
                return fileContents;
            }
        };
    }


    @Test
    public void empty_file() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(0, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(1, analyzer.getEmptyLines());
    }

    @Test
    public void only_whitespace() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer(" \t");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(0, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(1, analyzer.getEmptyLines());
    }

    @Test
    public void one_newline() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("\n");
        assertEquals(2, analyzer.getTotalLines());
        assertEquals(0, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(2, analyzer.getEmptyLines());
    }

    @Test
    public void one_code_line() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("some code");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(1, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }


    @Test
    public void one_end_of_line_comment() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("// a comment");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(0, analyzer.getCodeLines());
        assertEquals(1, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }

    @Test
    public void one_block_comment() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("/* a comment */");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(0, analyzer.getCodeLines());
        assertEquals(1, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }

    @Test
    public void code_and_end_of_line_comment() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("some code // a comment");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(1, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }

    @Test
    public void code_and_block_comment() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("some code /* a comment */");
        assertEquals(1, analyzer.getTotalLines());
        assertEquals(1, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }

    @Test
    public void code_and_multiline_block_comment() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("some code /* a multiline \n comment */");
        assertEquals(2, analyzer.getTotalLines());
        assertEquals(1, analyzer.getCodeLines());
        assertEquals(1, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }

    @Test
    public void code_and_multiline_block_comment_and_more_code() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("some code /* a multiline \n comment */ more code");
        assertEquals(2, analyzer.getTotalLines());
        assertEquals(2, analyzer.getCodeLines());
        assertEquals(0, analyzer.getCommentLines());
        assertEquals(0, analyzer.getEmptyLines());
    }

    @Test
    public void mixed_code_and_comments_and_empty_lines() throws IOException {
        FileAnalyzer analyzer = createFileAnalyzer("" +
                "public class HelloWorld {\n" +
                "    \n" +
                "    /**\n" +
                "     * This program will print \"Hello world\"\n" +
                "     */\n" +
                "    public static void main(String[] args) {\n" +
                "        System.out.println(\"Hello world\");  // print \"Hello world\"\n" +
                "    }\n" +
                "}");
        assertEquals(9, analyzer.getTotalLines());
        assertEquals(5, analyzer.getCodeLines());
        assertEquals(3, analyzer.getCommentLines());
        assertEquals(1, analyzer.getEmptyLines());
    }
}
