// Copyright © 2005-2009, Esko Luontola <www.orfjackal.net>
// This software is released under the Apache License 2.0.
// The license text is at http://www.apache.org/licenses/LICENSE-2.0

package net.orfjackal.sorsastus;

import org.junit.*;

import java.io.*;
import java.net.*;
import java.util.Scanner;

/**
 * @author Esko Luontola
 */
public class SorsastusIntegrationTest extends Assert {

    private PrintStream originalSystemOut;
    private ByteArrayOutputStream fakeSystemOut;

    @Before
    public void captureSystemOut() {
        originalSystemOut = System.out;
        fakeSystemOut = new ByteArrayOutputStream();
        System.setOut(new PrintStream(fakeSystemOut));
    }

    @After
    public void uncaptureSystemOut() {
        System.setOut(originalSystemOut);
    }

    public String getCapturedSystemOut() {
        Scanner sc = new Scanner(new ByteArrayInputStream(fakeSystemOut.toByteArray()));
        StringBuilder sb = new StringBuilder();
        while (sc.hasNextLine()) {
            sb.append(sc.nextLine()).append('\n');
        }
        return sb.toString();
    }

    private File getResourceAsFile(String path) {
        try {
            URL resource = getClass().getResource(path);
            File file = new File(resource.toURI());
            assertTrue(file.exists());
            return file;
        } catch (URISyntaxException e) {
            throw new RuntimeException(e);
        }
    }

    private void programShouldPrint(String expected) {
        assertEquals(expected, getCapturedSystemOut());
    }


    @Test
    public void testdata_ex1_code_lines() {
        File file = getResourceAsFile("/testdata/ex1/testData/TestSourceFile1.java");

        Sorsastus.main(new String[]{"--code-lines", file.getPath()});

        programShouldPrint("" +
                "5\t" + file.getPath() + "\n" +
                "5\ttotal\n");
    }

    @Test
    public void testdata_ex1_comment_lines() {
        File file = getResourceAsFile("/testdata/ex1/testData/TestSourceFile1.java");

        Sorsastus.main(new String[]{"--comment-lines", file.getPath()});

        programShouldPrint("" +
                "12\t" + file.getPath() + "\n" +
                "12\ttotal\n");
    }

    @Test
    public void testdata_ex1_empty_lines() {
        File file = getResourceAsFile("/testdata/ex1/testData/TestSourceFile1.java");

        Sorsastus.main(new String[]{"--empty-lines", file.getPath()});

        programShouldPrint("" +
                "3\t" + file.getPath() + "\n" +
                "3\ttotal\n");
    }

    @Test
    public void testdata_ex1_total_lines() {
        File file = getResourceAsFile("/testdata/ex1/testData/TestSourceFile1.java");

        Sorsastus.main(new String[]{"--total-lines", file.getPath()});

        programShouldPrint("" +
                "20\t" + file.getPath() + "\n" +
                "20\ttotal\n");
    }

    @Test
    public void testdata_ex2_structure() {
        File dir = getResourceAsFile("/testdata/ex2/testData/");

        Sorsastus.main(new String[]{"--structure", dir.getPath()});

        programShouldPrint("" +
                new File(dir, "CodeAnalyzer.java").getPath() + "\n" +
                "    Package:\n" +
                "\tnet.orfjackal.sorsastus\n" +
                "    Imports:\n" +
                "\tantlr.ANTLRException\n" +
                "\torg.jetbrains.annotations.NotNull\n" +
                "\tjava.io.File\n" +
                "\tjava.io.IOException\n" +
                "\tjava.util.ArrayList\n" +
                "\tjava.util.List\n" +
                "    Classes:\n" +
                "\tCodeAnalyzer\n" +
                "\t    Methods (total of 11):\n" +
                "\t\tpublic CodeAnalyzer()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic void addFile(File file)\n" +
                "\t\t    Complexity: 4\t(Conditionals: 4, Loops: 0)\n" +
                "\t\tpublic boolean isCountCodeLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic void setCountCodeLines(boolean countCodeLines)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic boolean isCountCommentLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic void setCountCommentLines(boolean countCommentLines)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic boolean isCountEmptyLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic void setCountEmptyLines(boolean countEmptyLines)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic String getLineCountReport()\n" +
                "\t\t    Complexity: 1\t(Conditionals: 0, Loops: 1)\n" +
                "\t\tprivate int getLineCount(FileAnalyzer file)\n" +
                "\t\t    Complexity: 3\t(Conditionals: 3, Loops: 0)\n" +
                "\t\tpublic String getStructureReport()\n" +
                "\t\t    Complexity: 1\t(Conditionals: 0, Loops: 1)\n" +
                "\n" +
                new File(dir, "FileAnalyzer.java").getPath() + "\n" +
                "    Package:\n" +
                "\tnet.orfjackal.sorsastus\n" +
                "    Imports:\n" +
                "\tantlr.ANTLRException\n" +
                "\tantlr.collections.AST\n" +
                "\torg.jetbrains.annotations.NotNull\n" +
                "\torg.jetbrains.annotations.Nullable\n" +
                "\tutil.java15grammar.JavaLexer\n" +
                "\tutil.java15grammar.JavaRecognizer\n" +
                "\tutil.java15grammar.JavaTreeParserTokenTypes\n" +
                "\tjava.io.*\n" +
                "\tjava.util.ArrayList\n" +
                "\tjava.util.List\n" +
                "    Classes:\n" +
                "\tFileAnalyzer\n" +
                "\t    Methods (total of 15):\n" +
                "\t\tpublic FileAnalyzer(File file)\n" +
                "\t\t    Complexity: 1\t(Conditionals: 1, Loops: 0)\n" +
                "\t\tprivate void countLines()\n" +
                "\t\t    Complexity: 14\t(Conditionals: 13, Loops: 1)\n" +
                "\t\tprivate AST getTree()\n" +
                "\t\t    Complexity: 1\t(Conditionals: 1, Loops: 0)\n" +
                "\t\tpublic String getStructureReport()\n" +
                "\t\t    Complexity: 5\t(Conditionals: 3, Loops: 2)\n" +
                "\t\tprivate static String getClassStructureReport(AST classNode)\n" +
                "\t\t    Complexity: 12\t(Conditionals: 7, Loops: 5)\n" +
                "\t\tprivate static void findNodes(AST node, int nodeType, List result)\n" +
                "\t\t    Complexity: 3\t(Conditionals: 2, Loops: 1)\n" +
                "\t\tprivate static String nodeToString(AST node)\n" +
                "\t\t    Complexity: 3\t(Conditionals: 2, Loops: 1)\n" +
                "\t\tprivate static int getNodeCount(AST node, int include, int exclude)\n" +
                "\t\t    Complexity: 5\t(Conditionals: 3, Loops: 2)\n" +
                "\t\tpublic File getFile()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic int getTotalLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic int getCodeLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic int getCommentLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic int getEmptyLines()\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tprivate static String readFile(File file)\n" +
                "\t\t    Complexity: 2\t(Conditionals: 1, Loops: 1)\n" +
                "\t\tpublic static void main(String[] args)\n" +
                "\t\t    Complexity: 1\t(Conditionals: 0, Loops: 1)\n" +
                "\n" +
                new File(dir, "Sorsastus.java").getPath() + "\n" +
                "    Package:\n" +
                "\tnet.orfjackal.sorsastus\n" +
                "    Classes:\n" +
                "\tSorsastus\n" +
                "\t    Methods (total of 1):\n" +
                "\t\tpublic static void main(String[] args)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\n" +
                new File(dir, "SorsastusLauncher.java").getPath() + "\n" +
                "    Package:\n" +
                "\tnet.orfjackal.sorsastus\n" +
                "    Imports:\n" +
                "\tnet.orfjackal.util.commandline.CommandLineParser\n" +
                "\tnet.orfjackal.util.commandline.FileArgument\n" +
                "\tnet.orfjackal.util.commandline.Option\n" +
                "\tjava.io.File\n" +
                "\tjava.io.IOException\n" +
                "\tjava.text.ParseException\n" +
                "\tantlr.ANTLRException\n" +
                "    Classes:\n" +
                "\tSorsastusLauncher\n" +
                "\t    Methods (total of 2):\n" +
                "\t\tpublic SorsastusLauncher(CodeAnalyzer analyzer)\n" +
                "\t\t    Complexity: 1\t(Conditionals: 1, Loops: 0)\n" +
                "\t\tpublic void parse(String[] params)\n" +
                "\t\t    Complexity: 2\t(Conditionals: 2, Loops: 0)\n" +
                "\n");
    }

    @Test
    public void testdata_ex3_structure_oo_metrics() {
        File file1 = getResourceAsFile("/testdata/ex3/testData/net/orfjackal/util/commandline/FileArgument.java");
        File file2 = getResourceAsFile("/testdata/ex3/testData/tests/InterfaceTest2.java");
        File file3 = getResourceAsFile("/testdata/ex3/testData/tests/InterfaceTest3b.java");

        Sorsastus.main(new String[]{"--structure", "--oo-metrics", file1.getPath(), file2.getPath(), file3.getPath()});

        programShouldPrint("" +
                file1.getPath() + "\n" +
                "    Package:\n" +
                "\tnet.orfjackal.util.commandline\n" +
                "    Imports:\n" +
                "\tnet.orfjackal.util.FileTools\n" +
                "\tjava.io.File\n" +
                "\tjava.text.ParseException\n" +
                "\tjava.util.ArrayList\n" +
                "\tjava.util.Arrays\n" +
                "\tjava.util.List\n" +
                "    Classes:\n" +
                "\tFileArgument\n" +
                "\t    WMC: 2\t(weighted methods per class)\n" +
                "\t    DIT: 1\t(depth of inheritance tree)\n" +
                "\t    NOC: 1\t(number of children)\n" +
                "\t    Methods (total of 4):\n" +
                "\t\tpublic FileArgument(String name, boolean required)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic FileArgument(String name, boolean required, boolean multiple)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\t\tpublic void onSpecified(String[] args)\n" +
                "\t\t    Complexity: 2\t(Conditionals: 1, Loops: 1)\n" +
                "\t\tpublic abstract void onSpecified(File[] files)\n" +
                "\t\t    Complexity: 0\t(Conditionals: 0, Loops: 0)\n" +
                "\n" +
                file2.getPath() + "\n" +
                "    Package:\n" +
                "\ttests\n" +
                "    Classes:\n" +
                "\tInterfaceTest2\n" +
                "\t    WMC: 0\t(weighted methods per class)\n" +
                "\t    DIT: 1\t(depth of inheritance tree)\n" +
                "\t    NOC: 2\t(number of children)\n" +
                "\t    Methods (total of 0):\n" +
                "\n" +
                file3.getPath() + "\n" +
                "    Package:\n" +
                "\ttests\n" +
                "    Classes:\n" +
                "\tInterfaceTest3b\n" +
                "\t    WMC: 0\t(weighted methods per class)\n" +
                "\t    DIT: 2\t(depth of inheritance tree)\n" +
                "\t    NOC: 0\t(number of children)\n" +
                "\t    Methods (total of 0):\n" +
                "\n");
    }
}
