package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Creates reports of the provided source code files.
 *
 * @author Esko Luontola
 */
public class CodeAnalyzer {

    @NotNull
    private List<FileAnalyzer> files;

    private boolean countCodeLines;

    private boolean countCommentLines;

    private boolean countEmptyLines;

    public CodeAnalyzer() {
        files = new ArrayList<FileAnalyzer>();
        countCodeLines = true;
        countCommentLines = false;
        countEmptyLines = false;
    }

    public void addFile(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException();
        } else if (!file.exists()) {
            throw new IOException("No such file: " + file);
        } else if (!file.isFile()) {
            throw new IOException("Not a file: " + file);
        } else if (!file.canRead()) {
            throw new IOException("File can not be read: " + file);
        }
        files.add(new FileAnalyzer(file));
    }

    public boolean isCountCodeLines() {
        return countCodeLines;
    }

    public void setCountCodeLines(boolean countCodeLines) {
        this.countCodeLines = countCodeLines;
    }

    public boolean isCountCommentLines() {
        return countCommentLines;
    }

    public void setCountCommentLines(boolean countCommentLines) {
        this.countCommentLines = countCommentLines;
    }

    public boolean isCountEmptyLines() {
        return countEmptyLines;
    }

    public void setCountEmptyLines(boolean countEmptyLines) {
        this.countEmptyLines = countEmptyLines;
    }

    /**
     * Returns a report which contains the line counts of each of the input files. For each file is reported how many
     * lines of code it includes (according to the set LOC definition) and the name of the file. The total sum of the
     * line counts is told on the last line.
     * <p/>
     * The report will end with the line freed character '\n'.
     */
    @NotNull
    public String getLineCountReport() {
        StringBuilder sb = new StringBuilder();
        int totalLines = 0;
        for (FileAnalyzer file : files) {
            int lines = getLineCount(file);
            totalLines += lines;
            sb.append(lines);
            sb.append('\t');
            sb.append(file.getFile().getPath());
            sb.append('\n');
        }
        sb.append(totalLines);
        sb.append("\ttotal\n");
        return sb.toString();
    }

    private int getLineCount(FileAnalyzer file) {
        int count = 0;
        if (isCountCodeLines()) {
            count += file.getCodeLines();
        }
        if (isCountCommentLines()) {
            count += file.getCommentLines();
        }
        if (isCountEmptyLines()) {
            count += file.getEmptyLines();
        }
        return count;
    }

    /**
     * Returns a report which contains a structural analysis of each of the input files. The report will contain
     * <ul><li>the name of the source code file(s) processed</li> <li>for each source file:</li> <ul><li>the name of the
     * package the file belongs to,</li><li>the number and names of packages imported,</li><li>the number and names of
     * classes in the file,</li></ul> <li>for each class:</li> <ul><li>the number and names of the methods in the
     * class,</li></ul> <li>for each method:</li> <ul><li>the number of if and switch statements,</li><li>the number of
     * for, while and do-while statements.</li></ul></ul> The report will end with the line freed character '\n'.
     *
     * @throws ANTLRException if there was a parse exception in processing one of the files.
     */
    @NotNull
    public String getStructureReport() throws ANTLRException {
        StringBuilder sb = new StringBuilder();
        for (FileAnalyzer file : files) {
            sb.append(file.getFile().getPath());
            sb.append('\n');
            try {
                sb.append(file.getStructureReport());
                sb.append('\n');
            } catch (ANTLRException e) {
                e.printStackTrace(); // DEBUG
                throw new ANTLRException("Parse exception in file '" + file.getFile() + "': " + e.getMessage());
            }
        }
        return sb.toString();
    }
}
