package net.orfjackal.sorsastus;

import antlr.ANTLRException;
import antlr.collections.AST;
import static net.orfjackal.sorsastus.FileAnalyzer.CursorPosition.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import util.java15grammar.JavaLexer;
import util.java15grammar.JavaRecognizer;
import util.java15grammar.JavaTreeParserTokenTypes;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

/**
 * Reads a source code file and collects statistical measures from it.
 *
 * @author Esko Luontola
 */
public class FileAnalyzer {

    /**
     * The file that is being analyzed.
     */
    @NotNull
    private File file;

    /**
     * The text content of the file.
     */
    @NotNull
    private String content;

    /**
     * Parsed AST tree of the source file. Is null, until the tree is created on request by calling getTree().
     */
    @Nullable
    private AST tree;

    /**
     * Number of lines in the file that contain code.
     */
    private int codeLines;

    /**
     * Number of lines in the file that do not containt code but contain comments.
     */
    private int commentLines;

    /**
     * Number of lines in the file that contain no code and no comments.
     */
    private int emptyLines;

    /**
     * Creates a new file analyzer and collects measures from the given file. Will read contents of the file to memory.
     *
     * @param file the file to be analyzed.
     * @throws IOException          if the file can not be read.
     * @throws NullPointerException if file is null.
     */
    public FileAnalyzer(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException();
        }
        this.file = file;
        this.content = readFile(file);
        countLines();
    }

    /**
     * Counts the code, comment and empty lines from the content of the analyzed file. Assumes that the content of the
     * file has already been read to the <code>this.content</code> field. The results will be saved to
     * <code>this.codeLines</code>, <code>this.commentLines</code> and <code>this.emptyLines</code>.
     */
    private void countLines() {
        this.codeLines = 0;
        this.commentLines = 0;
        this.emptyLines = 0;

        CursorPosition currentPos = CODE_BLOCK;
        boolean lineContainsCode = false;
        boolean lineContainsComments = false;

        for (int i = 0; i <= content.length(); i++) {
            char c;
            if (i < content.length()) {
                c = content.charAt(i);
            } else {
                c = '\n';                                   // treat End Of File the same way as End Of Line
            }

            // locate the end of line: '\n' or EOF
            if (c == '\n') {
                switch (currentPos) {
                case CODE_BLOCK:
                    if (lineContainsCode) {                 // this line contained some code, maybe also comments
                        codeLines++;
                    } else if (lineContainsComments) {      // comment block ended on this line
                        commentLines++;
                    } else {                                // no code, no comments -> empty line
                        emptyLines++;
                    }
                    lineContainsComments = false;
                    break;

                case COMMENT_BLOCK:
                    if (lineContainsCode) {
                        codeLines++;
                    } else {
                        commentLines++;
                    }
                    lineContainsComments = true;            // comment block continues to the next line
                    break;

                case COMMENT_LINE:
                    if (lineContainsCode) {
                        codeLines++;
                    } else {
                        commentLines++;
                    }
                    currentPos = CODE_BLOCK;                // next line will be again a normal code line
                    lineContainsComments = false;
                    break;

                default:
                    assert false;
                }

                // begin analyzing the next line
                lineContainsCode = false;
                continue;
            }

            // determine where comments begin and end
            switch (currentPos) {
            case CODE_BLOCK:

                // check if here is the beginning of a comment: "/*" or "//"
                if (c == '/') {
                    char c2 = content.charAt(i + 1);
                    if (c2 == '*') {                        // beginning of a comment block
                        currentPos = COMMENT_BLOCK;
                        lineContainsComments = true;
                        i++;                                // skip analyzing the '*' character
                        continue;
                    } else if (c2 == '/') {                 // beginning of a comment line
                        currentPos = COMMENT_LINE;
                        lineContainsComments = true;
                        i++;                                // skip analyzing the '/' character
                        continue;
                    }
                }

                // check if this line contains some code, so that it is not an empty line
                if (!Character.isWhitespace(c)) {
                    lineContainsCode = true;
                }
                break;

            case COMMENT_BLOCK:

                // check if here is the end of the comment block: "*/"
                if (c == '*' && content.charAt(i + 1) == '/') {
                    currentPos = CODE_BLOCK;
                    i++;                                    // skip analyzing the '/' character
                    continue;
                }
                break;

            case COMMENT_LINE:
                // NOOP: just wait for a new line to begin
                break;

            default:
                assert false;
            }
        }
    }

    /**
     * Returns the AST tree of the source file. The tree will be created on the first call of this method. After that
     * will be returned just a reference to the tree that was created previously.
     *
     * @return the AST tree of the source file.
     * @throws IllegalArgumentException if there was a parse exception in parsing the source file.
     */
    @NotNull
    private AST getTree() throws ANTLRException {
        if (tree == null) {

            // Create a scanner that reads from the input stream passed to us
            JavaLexer lexer = new JavaLexer(new StringReader(content));
            lexer.setFilename(file.getName());

            // Create a parser that reads from the scanner
            JavaRecognizer parser = new JavaRecognizer(lexer);
            parser.setFilename(file.getName());

            // start parsing at the compilationUnit rule
            parser.compilationUnit();
            tree = parser.getAST();
            assert tree != null;
        }
        return tree;
    }

    /**
     * Returns a report which contains a structural analysis of this file. The report will contain <ul><li>the name of
     * the package the file belongs to,</li><li>the number and names of packages imported,</li><li>the number and names
     * of classes in the file,</li></ul> <li>for each class:</li> <ul><li>the number and names of the methods in the
     * class,</li></ul> <li>for each method:</li> <ul><li>the number of if and switch statements,</li><li>the number of
     * for, while and do-while statements.</li></ul></ul> The report will end with the line freed character '\n'.
     */
    @NotNull
    public String getStructureReport() throws ANTLRException {
        StringBuilder sb = new StringBuilder();
        AST root = getTree();
//        System.out.println(root.toStringList()); // DEBUG

//        System.err.println("package:");
        List<AST> packageNodes = new ArrayList<AST>();
        findNodes(root, JavaTreeParserTokenTypes.PACKAGE_DEF, packageNodes);

//        System.err.println("imports:");
        List<AST> importNodes = new ArrayList<AST>();
        findNodes(root, JavaTreeParserTokenTypes.IMPORT, importNodes);

//        System.err.println("classes:");
        List<AST> classNodes = new ArrayList<AST>();
        findNodes(root, JavaTreeParserTokenTypes.CLASS_DEF, classNodes);

//        System.err.println("anonymous classes");
        List<AST> anonymClassNodes = new ArrayList<AST>();
        findNodes(root, JavaTreeParserTokenTypes.OBJBLOCK, anonymClassNodes);

        // package report
        if (packageNodes.size() > 0) {
            sb.append("    Package:\n");
            sb.append('\t').append(nodeToString(packageNodes.get(0).getFirstChild().getNextSibling())).append('\n');
        }

        // imports report
        if (importNodes.size() > 0) {
            sb.append("    Imports:\n");
            for (AST node : importNodes) {
                sb.append('\t').append(nodeToString(node.getFirstChild())).append('\n');
            }
        }

        // classes report
        if (classNodes.size() + anonymClassNodes.size() > 0) {
            sb.append("    Classes:\n");
            for (AST node : classNodes) {
                sb.append(getClassStructureReport(node));
            }
            // TODO: is a report of anonymous classes also needed?
//            for (AST node : anonymClassNodes) {
//                sb.append(getClassStructureReport(node));
//            }
        }
        return sb.toString();
    }

    private static String getClassStructureReport(AST classNode) {
        if (classNode.getType() != JavaTreeParserTokenTypes.CLASS_DEF) {
            throw new IllegalArgumentException("classNode.getType() = " + classNode.getType());
        }
        StringBuilder sb = new StringBuilder();

        // components of the class definition
        @NotNull AST cModifiers = null;
        @NotNull AST cIdent = null;
        @NotNull AST cExtendsClause = null;
        @NotNull AST cImplementsClause = null;
        @NotNull AST cObjBlock = null;

        // find the components of the class definition
        AST n = classNode.getFirstChild();
        while (n != null) {
            switch (n.getType()) {
            case JavaTreeParserTokenTypes.MODIFIERS:
                cModifiers = n;
                break;
            case JavaTreeParserTokenTypes.IDENT:
                cIdent = n;
                break;
            case JavaTreeParserTokenTypes.EXTENDS_CLAUSE:
                cExtendsClause = n;
                break;
            case JavaTreeParserTokenTypes.IMPLEMENTS_CLAUSE:
                cImplementsClause = n;
                break;
            case JavaTreeParserTokenTypes.OBJBLOCK:
                cObjBlock = n;
                break;
            default:
                assert false : "n.getType() = " + n.getType();
                break;
            }
            n = n.getNextSibling();
        }
        assert cModifiers != null;
        assert cIdent != null;
        assert cExtendsClause != null;
        assert cImplementsClause != null;
        assert cObjBlock != null;

        // name of the class
        sb.append('\t').append(cIdent.getText()).append('\n');

        // methods of the class
        List<AST> methodNodes = new ArrayList<AST>();
        findNodes(classNode, JavaTreeParserTokenTypes.CTOR_DEF, methodNodes);
        findNodes(classNode, JavaTreeParserTokenTypes.METHOD_DEF, methodNodes);
        // TODO: avoid the counting of methods of anonymous and private classes

        sb.append("\t    Methods (total of ").append(methodNodes.size()).append("):\n");
        for (AST methodNode : methodNodes) {

            // components of the method definition
            @NotNull AST mModifiers = null;
            @Nullable AST mType = null;        // null for constructors
            @NotNull AST mIdent = null;
            @NotNull AST mParameters = null;
            @NotNull AST mSlist = null;
            @Nullable AST mThrows = null;

            // find the components of the method definition
            AST m = methodNode.getFirstChild();
            while (m != null) {
                switch (m.getType()) {
                case JavaTreeParserTokenTypes.MODIFIERS:
                    mModifiers = m;
                    break;
                case JavaTreeParserTokenTypes.TYPE:
                    mType = m;
                    break;
                case JavaTreeParserTokenTypes.IDENT:
                    mIdent = m;
                    break;
                case JavaTreeParserTokenTypes.PARAMETERS:
                    mParameters = m;
                    break;
                case JavaTreeParserTokenTypes.SLIST:
                    mSlist = m;
                    break;
                case JavaTreeParserTokenTypes.LITERAL_throws:
                    mThrows = m;
                    break;
                default:
                    System.err.println(m.toStringTree());
                    assert false : "m.getType() = " + m.getType();
                    break;
                }
                m = m.getNextSibling();
            }
            assert mModifiers != null;
            assert mIdent != null;
            assert mParameters != null;
            assert mSlist != null;

            // begin printing the method definition
            sb.append("\t\t");

            // print modifiers
            m = mModifiers.getFirstChild();
            while (m != null) {
                if (m.getType() != JavaTreeParserTokenTypes.ANNOTATION) {
                    sb.append(m.getText()).append(' ');
                }
                m = m.getNextSibling();
            }

            // print type
            if (mType != null) {
                sb.append(mType.getFirstChild().getText()).append(' ');
            }

            // print method identifier
            sb.append(mIdent.getText()).append('(');
            m = mParameters.getFirstChild();
            while (m != null) {
                AST child = m.getFirstChild();                  // MODIFIERS
                AST pType = child = child.getNextSibling();     // TYPE
                AST pIdent = child.getNextSibling();            // parameter name

                if (pType.getFirstChild().getType() == JavaTreeParserTokenTypes.ARRAY_DECLARATOR) {
                    sb.append(pType.getFirstChild().getFirstChild().getText()).append("[]");
                } else {
                    sb.append(pType.getFirstChild().getText());
                }
                sb.append(' ');
                sb.append(pIdent.getText());

                m = m.getNextSibling();
                if (m != null) {
                    sb.append(", ");
                }
            }
            sb.append(')');
            sb.append('\n');

            // print the complexity of the method
            int ifCount = getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_if,
                    JavaTreeParserTokenTypes.OBJBLOCK);         // do not advance into anonymous inner classes
            int switchCount = getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_switch,
                    JavaTreeParserTokenTypes.OBJBLOCK);
            int forCount = getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_for,
                    JavaTreeParserTokenTypes.OBJBLOCK);
            int whileCount = getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_while,
                    JavaTreeParserTokenTypes.OBJBLOCK);         // does NOT count the do-while structures
            int doWhileCount = getNodeCount(mSlist, JavaTreeParserTokenTypes.LITERAL_do,
                    JavaTreeParserTokenTypes.OBJBLOCK);

            sb.append("\t\t    Conditionals: ").append(ifCount + switchCount);
            sb.append("\tLoops: ").append(forCount + whileCount + doWhileCount).append('\n');
        }

        return sb.toString();
    }

    /**
     * Finds from the provided AST tree all the nodes of the type nodeType and adds them to the result list.
     *
     * @param node     the root node of the tree from where to search. If null, will do nothing.
     * @param nodeType the type of the nodes to be searched for.
     * @param result   the nodes that were found will be added to this list.
     * @throws NullPointerException if result is null and a matching node was found.
     */
    private static void findNodes(AST node, int nodeType, List<AST> result) {
        while (node != null) {
            if (node.getNumberOfChildren() > 0) {
                findNodes(node.getFirstChild(), nodeType, result);
            }
            if (node.getType() == nodeType) {
                //System.out.println(node.toStringTree()); // DEBUG
                result.add(node);
            }
            node = node.getNextSibling();
        }
    }

    private static String nodeToString(AST node) {
        if (node == null) {
            return "";
        }
        if (node.getNumberOfChildren() > 0) {
            StringBuilder sb = new StringBuilder();

            AST child = node.getFirstChild();
            sb.append(nodeToString(child));
            sb.append(node.getText());

            do {
                child = child.getNextSibling();
                sb.append(nodeToString(child));
            } while (child != null);

//            sb.append("["+nodeToString(node.getNextSibling())+"]");
            return sb.toString();
        } else {
            return node.getText();
        }
    }

    /**
     * Returns the count of the nodes of type 'include' in the AST tree whose root is 'node'. Does not advance into
     * nodes of the type 'exclude'.
     *
     * @param node    the root node where to begin the search.
     * @param include the type of the nodes to count.
     * @param exclude the types of the nodes into which the algorithm should not dig into.
     * @return the total count of nodes of type 'include' under 'node'.
     */
    private static int getNodeCount(AST node, int include, int... exclude) {
        if (node == null) {
            return 0;
        }
        for (int ex : exclude) {
            if (node.getType() == ex) {
                return 0;
            }
        }
        int count = 0;
        AST child = node.getFirstChild();
        while (child != null) {
            count += getNodeCount(child, include, exclude);
            child = child.getNextSibling();
        }
        if (node.getType() == include) {
            count++;
        }
        return count;
    }

    /**
     * Returns the file that was analyzed.
     */
    public File getFile() {
        return file;
    }

    /**
     * Returns the total number of lines in the file, including all comments and empty lines.
     */
    public int getTotalLines() {
        return codeLines + commentLines + emptyLines;
    }

    /**
     * Returns the number of code lines in the file, excluding comment lines and empty lines.
     */
    public int getCodeLines() {
        return codeLines;
    }

    /**
     * Returns the number of lines in the file that contain only comments and no code.
     */
    public int getCommentLines() {
        return commentLines;
    }

    /**
     * Returns the number of lines that contain no code and no comments.
     */
    public int getEmptyLines() {
        return emptyLines;
    }

    /**
     * Reads the given file to a string.
     *
     * @param file the file to be read.
     * @return the contents of the file as a string.
     * @throws IOException          if the file can not be read.
     * @throws NullPointerException if file is null.
     */
    @NotNull
    private static String readFile(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException();
        }
        Reader in = new FileReader(file);
        StringBuilder sb = new StringBuilder((int) file.length());
        int i;
        while ((i = in.read()) != -1) {
            sb.append((char) i);
        }
        in.close();
        return sb.toString();
    }

    protected enum CursorPosition {
        CODE_BLOCK, COMMENT_BLOCK, COMMENT_LINE
    }

    public static void main(String[] args) throws IOException {
        File[] files = new File[]{
                new File("tests/TestSourceFile1.java")
        };
        for (File file : files) {
            FileAnalyzer a = new FileAnalyzer(file);
            System.out.println(a.getFile());
            System.out.println("Total lines:    " + a.getTotalLines());
            System.out.println("Comment lines:  " + a.getCommentLines());
            System.out.println("Code lines:     " + a.getCodeLines());
            System.out.println("Empty lines:    " + a.getEmptyLines());
            System.out.println();
        }
    }
}
