package net.orfjackal.sorsastus;

import antlr.collections.AST;
import net.orfjackal.util.ASTTools;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import util.java15grammar.JavaTreeParserTokenTypes;

import java.util.ArrayList;
import java.util.List;

/**
 * Collects metrics from the structure of a class.
 *
 * @author Esko Luontola
 */
public class ClassAnalyzer {

    @NotNull
    private FileAnalyzer fileAnalyzer;

    @NotNull
    private AST classNode;

    /* Analyzed data from this class. Initialized by analyzeStructure() */
    @NotNull
    private String className;
    @Nullable
    private String classExtends;
    @NotNull
    private String[] classImplements;
    @NotNull
    private MethodAnalyzer[] classMethods;

    /**
     * Creates a new ClassAnalyzer and analyzes the structure of the class. Is not (yet) able to analyze anonymous
     * classes.
     *
     * @param fileAnalyzer the file that this class belongs to.
     * @param classNode    structure of the class. The type of the node must be CLASS_DEF or INTERFACE_DEF.
     * @throws NullPointerException     if any parameter is null.
     * @throws IllegalArgumentException if classNode is of wrong type.
     */
    public ClassAnalyzer(FileAnalyzer fileAnalyzer, AST classNode) {
        if (fileAnalyzer == null) {
            throw new NullPointerException();
        } else if (classNode == null) {
            throw new NullPointerException();
        } else if (classNode.getType() != JavaTreeParserTokenTypes.CLASS_DEF
                && classNode.getType() != JavaTreeParserTokenTypes.INTERFACE_DEF) {
            throw new IllegalArgumentException("classNode.getType() = " + classNode.getType());
        }
        this.fileAnalyzer = fileAnalyzer;
        this.classNode = classNode;
        analyzeStructure();
    }

    /**
     * Analyzes the AST tree of this class and saves the results for later use.
     */
    private void analyzeStructure() {

        // components of the class definition
        @NotNull AST cModifiers = null;
        @NotNull AST cIdent = null;
        @NotNull AST cExtendsClause = null;
        @Nullable AST cImplementsClause = null;     // null if interface class
        @NotNull AST cObjBlock = null;

        // find the components of the class definition
        AST n = classNode.getFirstChild();
        while (n != null) {
            switch (n.getType()) {
            case JavaTreeParserTokenTypes.MODIFIERS:
                cModifiers = n;
                break;
            case JavaTreeParserTokenTypes.IDENT:
                cIdent = n;
                break;
            case JavaTreeParserTokenTypes.EXTENDS_CLAUSE:
                cExtendsClause = n;
                break;
            case JavaTreeParserTokenTypes.IMPLEMENTS_CLAUSE:
                cImplementsClause = n;
                break;
            case JavaTreeParserTokenTypes.OBJBLOCK:
                cObjBlock = n;
                break;
            default:
                assert false : "n.getType() = " + n.getType();
                break;
            }
            n = n.getNextSibling();
        }
        assert cModifiers != null;
        assert cIdent != null;
        assert cExtendsClause != null;
        assert cObjBlock != null;

        // name of the class
        this.className = cIdent.getText();

        // classes extended by the class
        if (cExtendsClause.getFirstChild() != null) {
            this.classExtends = ASTTools.nodeToString(cExtendsClause.getFirstChild());
        } else {
            this.classExtends = null;
        }

        // interfaces implemented by the class
        List<String> interfaces = new ArrayList<String>();
        if (cImplementsClause != null) {
            n = cImplementsClause.getFirstChild();
            while (n != null) {
                interfaces.add(ASTTools.nodeToString(n));
                n = n.getNextSibling();
            }
        }
        this.classImplements = interfaces.toArray(new String[interfaces.size()]);

        // methods of the class
        List<AST> methodNodes = new ArrayList<AST>();
        ASTTools.findNodes(cObjBlock, JavaTreeParserTokenTypes.CTOR_DEF, methodNodes);
        ASTTools.findNodes(cObjBlock, JavaTreeParserTokenTypes.METHOD_DEF, methodNodes);
        // TODO: avoid the counting of methods of anonymous and private classes

        List<MethodAnalyzer> list = new ArrayList<MethodAnalyzer>(methodNodes.size());
        for (AST node : methodNodes) {
            list.add(new MethodAnalyzer(this, node));
        }
        this.classMethods = list.toArray(new MethodAnalyzer[list.size()]);
    }

    /**
     * Returns a detailed report of the structure of this class. The returned string will end with the line feed
     * character '\n'.
     */
    @NotNull
    public String getStructureReport() {
        StringBuilder sb = new StringBuilder();

        // name of the class
        sb.append('\t').append(className).append('\n');

        // object-oriented metrics of the class
        if (fileAnalyzer.getAnalyzerBatch().isIncludeOOMetrics()) {
            sb.append("\t    WMC: ").append(getWMC()).append("\t(weighted methods per class)\n");
            sb.append("\t    DIT: ").append(getDIT()).append("\t(depth of inheritance tree)\n");
            sb.append("\t    NOC: ").append(getNOC()).append("\t(number of children)\n");
        }

        // methods of the class
        sb.append("\t    Methods (total of ").append(classMethods.length).append("):\n");
        for (MethodAnalyzer method : classMethods) {
            sb.append("\t\t").append(method.getDefinition());
            sb.append('\n');
            sb.append("\t\t    Complexity: ").append(method.getComplexity());
            sb.append("\t(Conditionals: ").append(method.getConditionals());
            sb.append(", Loops: ").append(method.getLoops()).append(")");
            sb.append('\n');
        }

        return sb.toString();
    }

    /**
     * Returns the name of this class.
     */
    @NotNull
    public String getName() {
        return className;
    }

    /**
     * Returns the name of the package that this class belongs to.
     */
    @NotNull
    public String getPackage() {
        return fileAnalyzer.getPackage();
    }

    /**
     * Returns the ClassAnalyzer of the class that this class extends, if it is in the classpath.
     *
     * @return the associated ClassAnalyzer, or null if this class does not extend anything or if the extended class was
     *         not found.
     */
    @Nullable
    public ClassAnalyzer getExtends() {
        if (classExtends != null) {
            return fileAnalyzer.findClass(classExtends);
        }
        return null;
    }

    /**
     * Returns the ClassAnalyzers of the classes that this class implements, if they are in the classpath.
     *
     * @return the associated ClassAnalyzers, or an empty array if this class does not implement anything or if the
     *         implemented classes were not found.
     */
    public ClassAnalyzer[] getImplements() {
        List<ClassAnalyzer> result = new ArrayList<ClassAnalyzer>();
        for (String s : classImplements) {
            ClassAnalyzer ca = fileAnalyzer.findClass(s);
            if (ca != null) {
                result.add(ca);
            }
        }
        return result.toArray(new ClassAnalyzer[result.size()]);
    }

    /**
     * Returns the "weighted methods per class" of this class. The weighted methods per class is the sum of the
     * complexities of the methods in the class. Here the complexity of a method is defined as the total number of
     * conditional structures and loop structures in the method.
     */
    public int getWMC() {
        int sum = 0;
        for (MethodAnalyzer method : classMethods) {
            sum += method.getComplexity();
        }
        return sum;
    }

    /**
     * Returns the "depth of inheritance tree" of this class. The depth of inheritance tree is the number of
     * superclasses of the class. Here only the depth of the inheritance tree seen from the source files via extends
     * statements is concidered. If a class extends a library class provided by the language, the class will be treated
     * as the root of the inheritance tree. Also the implements statement are considered as inheritance (in which case
     * the largest depth of the inheritance tree is selected as the DIT of the class).
     */
    public int getDIT() {
        ClassAnalyzer ca;
        int max = 0;
        if (classExtends != null) {
            ca = fileAnalyzer.findClass(classExtends);
            if (ca != null) {
                max = Math.max(max, ca.getDIT() + 1);
            }
        }
        for (String s : classImplements) {
            ca = fileAnalyzer.findClass(s);
            if (ca != null) {
                max = Math.max(max, ca.getDIT() + 1);
            }
        }
        return max;
    }

    /**
     * Returns the "number of children" of this class. The number of children is the number of subclasses directly
     * extending or implementing the class.
     */
    public int getNOC() {
        ClassAnalyzer[] children = fileAnalyzer.getAnalyzerBatch().findChildClasses(this);
        return children.length;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj != null && obj instanceof ClassAnalyzer) {
            ClassAnalyzer other = (ClassAnalyzer) obj;
            return this.getPackage().equals(other.getPackage()) && this.getName().equals(other.getName());
        }
        return false;
    }
}
