// Tietorakenteiden harjoitusty (syksy 2004), Esko Luontola
// Tietoliikenneverkon reititysalgoritmien vertailu, "Sim Router"

/**
 * Verkossa kulkeva datapaketti, jolla on lht- ja ptepisteet.
 */
public class Packet implements Cloneable {
	
	// Status:
	public static final int TRANSFER         = 0;
	public static final int RECIEVED         = 1;
	public static final int DROPPED          = 2;
	public static final int DROPPED_NO_TTL   = 3;
	public static final int DROPPED_NO_ROUTE = 4;
	
	// Protocoll:
	public static final int ROUTE_MAP        = 0;
	public static final int FLOOD            = 1;
	public static final int HOT_POTATO       = 2;
	public static final int SHORTEST_ROUTE   = 3;
	public static final int FASTEST_ROUTE    = 4;
	
	private int status;
	
	private int protocoll;
	
	private int from;
	
	private int to;
	
	private int ttl;
	
	private int jumps;
	
	private int startTime;
	
	private int endTime;
	
	private int effectiveTransferTime;
	
	/**
	 * Luo paketin, joka matkaa from:sta to:hon kytten reititysalgoritmia
	 * protocoll. Paketti pudotetaan ttl hypyn jlkeen, paitsi jos ttl=0.
	 */
	public Packet(int from, int to, int protocoll, int ttl) {
		this.status = TRANSFER;
		this.protocoll = protocoll;
		this.from = from;
		this.to = to;
		this.ttl = ttl;
		this.jumps = 0;
		this.startTime = Ticker.getTime();
		this.endTime = 0;
		this.effectiveTransferTime = 0;
	}
	
	/**
	 * Luo kopion tst paketista.
	 */
	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
			return null;
		}
	}
	
	/**
	 * Lhettjn osoite.
	 */
	public int getFrom() {
		return from;
	}
	
	/**
	 * Vastaanottajan osoite.
	 */
	public int getTo() {
		return to;
	}
	
	/**
	 * Reitittmiseen kytettv algoritmi.
	 */
	public int getProtocoll() {
		return protocoll;
	}
	
	/**
	 * Kertoo onko paketilla viel elinaikaa. Jos false, niin reitittimen on
	 * poistettava paketti.
	 */
	public boolean hasTTL() {
		return (ttl == 0 || jumps < ttl);
	}
	
	/**
	 * Kasvattaa aktiivista siirtoaikaa annetulla ajalla ja hyppylaskuria yhdell.
	 */
	public void addTransferTime(int duration) {
		jumps++;
		effectiveTransferTime += duration;
	}
	
	/**
	 * Palauttaa ajan paketin lhettmisen ja saapumisen vlill, mukaanlukien 
	 * odotukset puskureissa.
	 */
	public int getTransferTime() {
		if (status == TRANSFER) {
			return (Ticker.getTime() - startTime);
		} else {
			return (endTime - startTime);
		}
	}
	
	/**
	 * Palauttaa aktiivisen siirtoajan, eli getTransferTime() poislukien
	 * odotukset puskureissa.
	 */
	public int getEffectiveTransferTime() {
		return effectiveTransferTime;
	}
	
	/**
	 * Hyppyjen lukumr.
	 */
	public int getJumps() {
		return jumps;
	}
	
	/**
	 * Paketin tmnhetkinen tila.
	 */
	public int getStatus() {
		return status;
	}
	
	/**
	 * Kuittaa paketin saapuneeksi mrnphns.
	 */
	public void markAsRecieved() {
		if (SimRouter.DEBUG) {
			System.out.println("Packet Recieved");
		}
		endTime = Ticker.getTime();
		status = RECIEVED;
		Statistics.registerPacket(this);
	}
	
	/**
	 * Kuittaa paketin pudotetuksi ruuhkan takia.
	 */
	public void markAsDropped() {
		if (SimRouter.DEBUG) {
			System.out.println("Packet Dropped");
		}
		endTime = Ticker.getTime();
		status = DROPPED;
		Statistics.registerPacket(this);
	}
	
	/**
	 * Kuittaa paketin pudotetuksi alinajan loputtua.
	 */
	public void markAsNoTTL() {
		if (SimRouter.DEBUG) {
			System.out.println("Packet Dropped (No TTL)");
		}
		endTime = Ticker.getTime();
		status = DROPPED_NO_TTL;
		Statistics.registerPacket(this);
	}
	
	/**
	 * Kuittaa paketin pudotetuksi, jos vastaanottajaan ei ole yhteytt.
	 */
	public void markAsNoRouteToHost() {
		if (SimRouter.DEBUG) {
			System.out.println("Packet Dropped (No Route To Host)");
		}
		endTime = Ticker.getTime();
		status = DROPPED_NO_ROUTE;
		Statistics.registerPacket(this);
	}
}
