// Tietorakenteiden harjoitusty (syksy 2004), Esko Luontola
// Tietoliikenneverkon reititysalgoritmien vertailu, "Sim Router"

/**
 * Reititin, joka ohjaa toisilta reitittimilt saamiaan paketteja eteenpin,
 * kunnes ne saapuvat mrnphns.
 */
public class Router {
	
	/**
	 * Reittittimen yksiliv osoite. Osoitteiden on alettava nollasta.
	 */
	private int id;
	
	/**
	 * DuplexPort-vektori, yhteydet toisiin reitittimiin.
	 */
	private DuplexPortVector ports;
	
	/**
	 * Reititystaulu: ReititinID -> (PorttiNro, Jumps). 
	 * 
	 * Kuinka monta hyppy kuhunkin reitittimeen on kunkin portin kautta. 
	 * Jos ei yhteytt, niin alkion arvo on -1. Reititystaulu on null, jos 
	 * reititystietoja ei ole saatu viimeisimmn yhteyden lismisen jlkeen.
	 */
	private int[][] shortestRoute;
	private static final int PORT_NUMBER = 0;
	private static final int JUMP_COUNT = 1;
	
	/**
	 * Reititystaulu: ReititinID -> PorttiNro -> EffectiveTransferTime.
	 * 
	 * Aktiivinen siirtoaika kuhunkin reitittimeen kunkin portin kautta.
	 * Jos ei yhteytt, niin alkion arvo on -1. Reititystaulu on null, jos 
	 * reititystietoja ei ole saatu viimeisimmn yhteyden lismisen jlkeen.
	 */
	private int[][] fastestRoute;
	
	/**
	 * Luo reitittimen, jonka osoite on id. Kutsuja on vastuussa, ettei kaksi
	 * reititint saa samaa osoitetta. Osoitteiden on alettava nollasta.
	 */
	public Router(int id) {
		if (id < 0) {
			throw new ArrayIndexOutOfBoundsException();
		}
		this.id = id;
		this.ports = new DuplexPortVector();
		this.resetRouteTables();
	}
	
	/**
	 * Palauttaa tmn reitittimen yksilivn osoiteen.
	 */
	public int getID() {
		return id;
	}
	
	/**
	 * Palauttaa tst reitittimest toisiin olevien yhteyksien lukumrn.
	 */
	public int getConnections() {
		return ports.size();
	}
	
	/**
	 * Tyhjent reitittimen lhetys- ja vastaanottopuskurit paketeista.
	 */
	public void flushPackets() {
		for (int i = 0; i < ports.size(); i++) {
			ports.get(i).input.flushPackets();
			ports.get(i).output.flushPackets();
		}
	}
	
	/**
	 * Poistaa reititystaulut.
	 */
	public void resetRouteTables() {
		shortestRoute = null;
		fastestRoute = null;
	}
	
	/**
	 * Alustaa reititystaulut.
	 */
	public void initRouteTables() {
		int routerCount = Simulation.getRouters();
		int portCount = ports.size();
		
		shortestRoute = new int[routerCount][2];
		fastestRoute = new int[routerCount][portCount];
		
		for (int i = 0; i < routerCount; i++) {
			shortestRoute[i][PORT_NUMBER] = -1;
			shortestRoute[i][JUMP_COUNT] = -1;
			for (int j = 0; j < portCount; j++) {
				fastestRoute[i][j] = -1;
			}
		}
	}
	
	/**
	 * Lhett reitityspaketin kaikkiin muihin yhdistettyihin reitittimiin.
	 * Jos Simulation.isOptimizeStack() on tosi, tytyy kyttjn itse kutsua 
	 * Ticker.tickUntilStopped() metodia. Muussa tapauksessa paketti siirtyy
	 * verkon lpi 0 aikayksikss.
	 */
	public void sendRouteMapPacket() {
		routePacket(new Packet(getID(), -1, Packet.ROUTE_MAP, 0));
	}
	
	/**
	 * Kertoo onko tst reitittimest mahdollista lhett toiseen reitittimeen.
	 */
	public boolean isRouteToHost(int other) {
		if (shortestRoute == null || shortestRoute[other][PORT_NUMBER] < 0) {
			return false;
		} else {
			return true;
		}
		
	}
	
	/**
	 * Yhdist tmn ja toisen reitittimen. Palauttaa true, jos yhdistys 
	 * onnistui (kummallakin alle maksimimr yhteyksi ja nm kaksi eivt 
	 * entuudestaan yhdistettyj), muutoin false.
	 */
	public boolean join(Router other) {
		return join(other, false);
	}
	
	/**
	 * Yhdist tmn ja toisen reitittimen. Jos force==true, yhdistys tehdn
	 * vaikka jompi kumpi reitittimist olisi jo saavuttanut maksimimrn 
	 * yhteyksi. Palauttaa true, jos yhdistys onnistui, muutoin false.
	 */
	public boolean join(Router other, boolean force) {
		// ei yhdistet itseens
		if (getID() == other.getID()) {
			return false;
		}
		// ei yhdistet kahdesti samaan
		for (int i = 0; i < ports.size(); i++) {
			if (ports.get(i).getTarget() == other.getID()) {
				return false;
			}
		}
		// ei ylitet maksimimr yhteyksi
		if (!force
			&& (ports.size() >= Simulation.getMaxPorts()
				|| other.ports.size() >= Simulation.getMaxPorts()
			)) {
			return false;
		}
		
		// tarkistukset ok, yhdist reitittimet
		RecievePort input1 = new RecievePort(this, Simulation.getBufferSize());
		RecievePort input2 = new RecievePort(other, Simulation.getBufferSize());
		SendPort output1 = new SendPort(this, input2,
		                                Simulation.getBufferSize(),
		                                Simulation.getRandomTransferDelay());
		SendPort output2 = new SendPort(other, input1,
		                                Simulation.getBufferSize(),
		                                Simulation.getRandomTransferDelay());
		DuplexPort port1 = new DuplexPort(input1, output1, other.getID());
		DuplexPort port2 = new DuplexPort(input2, output2, this.getID());
		
		this.ports.add(port1);
		other.ports.add(port2);
		
		// nollaa reititystaulut
		this.resetRouteTables();
		other.resetRouteTables();
		
		return true;
	}
	
	/**
	 * Vastaanottoportit kutsuvat tt metodia aina, kun ne vastaanottavat
	 * uuden paketin. Tm metodi ky kaikki vastaanottoportit lpi ja 
	 * ottaa jokaisesta yhden paketin, jonka sitten ohjaa eteenpin.
	 */
	public void tick() {
		for (int i = 0; i < ports.size(); i++) {
			if (!ports.get(i).input.isEmpty()) {
				Packet packet = ports.get(i).input.pop();
				routePacket(packet, i);
			}
		}
	}
	
	/**
	 * Reititt paketin, jonka lhtpiste on tm reititin.
	 */
	public void routePacket(Packet packet) {
		routePacket(packet, -1);
	}
	
	/**
	 * Reititt paketin, joka tuli portista fromPort. Porttinumero voi olla -1,
	 * jos tm reititin on paketin lhtpiste.
	 */
	public void routePacket(Packet packet, int fromPort) {
		if (packet.getTo() == getID()) {
			packet.markAsRecieved();
			return;
		}
		if (!packet.hasTTL()) {
			packet.markAsNoTTL();
			return;
		}
		
		switch (packet.getProtocoll()) {
			case Packet.ROUTE_MAP:
				routePacketRouteMap(packet, fromPort);
				break;
				
			case Packet.FLOOD:
				routePacketFlood(packet, fromPort);
				break;
				
			case Packet.HOT_POTATO:
				routePacketHotPotato(packet, fromPort);
				break;
				
			case Packet.SHORTEST_ROUTE:
				routePacketShortestRoute(packet, fromPort);
				break;
				
			case Packet.FASTEST_ROUTE:
				routePacketFastestRoute(packet, fromPort);
				break;
				
			default:
				System.out.println("Unknown protocoll "+packet.getProtocoll()
					+", packet dropped.");
				packet.markAsDropped();
				break;
		}
	}
	
	/**
	 * Reititystaulualgoritmi.
	 * 
	 * Tarkistaa reititystauluista tmn paketin lhettjn kohdalta, onko
	 * paketti saapunut lyhyemp reitti kuin reititystaulujen aiemmat tiedot.
	 * Jos on, niin pivitt reititystauluja ja tulvittaa paketin eteenpin.
	 * Muutoin ei tee mitn.
	 */
	private void routePacketRouteMap(Packet packet, int fromPort) {
		if (shortestRoute == null || fastestRoute == null) {
			initRouteTables();
		}
		
		boolean doForward = false;
		int fromRouter = packet.getFrom();
		
		if (fromRouter == getID()) {
			// lhetetty tst reitittimest
			doForward = true;
		} else {
			// saapunut muualta, tarkista saapuiko lyhyemp reitti
			if (shortestRoute[fromRouter][JUMP_COUNT] > packet.getJumps()
				|| shortestRoute[fromRouter][JUMP_COUNT] < 0) {
				shortestRoute[fromRouter][PORT_NUMBER] = fromPort;
				shortestRoute[fromRouter][JUMP_COUNT] = packet.getJumps();
				doForward = true;
			}
			if (fastestRoute[fromRouter][fromPort] > packet.getEffectiveTransferTime()
				|| fastestRoute[fromRouter][fromPort] < 0) {
				fastestRoute[fromRouter][fromPort] = packet.getEffectiveTransferTime();
				doForward = true;
			}
		}
		
		if (doForward) {
			routePacketFlood(packet, fromPort);
		}
	}
	
	/**
	 * Tulvitusalgoritmi.
	 * 
	 * Lhett kopion paketista kaikkiin portteihin paitsi sinne mist tuli.
	 * Jos muita portteja ei ole, ei tee mitn.
	 */
	private void routePacketFlood(Packet packet, int fromPort) {
		boolean wasSent = false;
		for (int i = 0; i < ports.size(); i++) {
			if (i != fromPort) {
				ports.get(i).output.push((Packet)packet.clone());
				wasSent = true;
			}
		}
		if (wasSent == false) {
			packet.markAsNoRouteToHost();
		}
	}
	
	/**
	 * Kuuma peruna -algoritmi.
	 * 
	 * Lhett paketin satunnaiseen porttiin. Voi olla mys sama kuin lhettj.
	 * Jos yhtn porttia ei ole, hylk paketin.
	 */
	private void routePacketHotPotato(Packet packet, int fromPort) {
		if (ports.size() > 0) {
			int toPort = (int)(Math.random() * ports.size());
			ports.get(toPort).output.push(packet);
		} else {
			packet.markAsNoRouteToHost();
		}
	}
	
	/**
	 * Lyhyimmn reitin algoritmi.
	 * 
	 * Tarkistaa shortestRoutePorts -reititystaulusta mink portin kautta on
	 * paketin mrnphn mahdollisimman vhn hyppyj ja lhett paketin
	 * sinne. Jos reititystaulusta ei lydy tietoa, paketti hyltn.
	 */
	private void routePacketShortestRoute(Packet packet, int fromPort) {
		if (shortestRoute == null
			|| shortestRoute[packet.getTo()][PORT_NUMBER] < 0) {
			System.out.println(shortestRoute);
			packet.markAsNoRouteToHost();
		} else {
			int toPort = shortestRoute[packet.getTo()][PORT_NUMBER];
			ports.get(toPort).output.push(packet);
		}
	}
	
	/**
	 * Nopeimman reitin algoritmi.
	 * 
	 * Tarkistaa fastestRouteDelays -reititystaulusta kuinka kauan kunkin portin
	 * kauttaa kest menn paketin mrnphn, ja lis lukuun kuinka kauan
	 * paketti joutuu odottamaan lhetyspuskurissa. Valitsee saaduista luvuista
	 * pienimmn ja lhett paketin kyseiseen porttiin. Jos reititystaulusta ei
	 * lydy tietoa, paketti hyltn.
	 */
	private void routePacketFastestRoute(Packet packet, int fromPort) {
		if (fastestRoute == null || ports.size() == 0) {
			packet.markAsNoRouteToHost();
		} else {
			int[] delays = fastestRoute[packet.getTo()];
			int fastestPort = -1;
			int fastestTime = -1;
			
			for (int i = 0; i < delays.length; i++) {
				if (delays[i] < 0) {
					// arvoa ei alustettu tai no route to host
					continue;
				}
				
				int realDelay;
				if (ports.get(i).output.isFull()) {
					realDelay = Integer.MAX_VALUE;
				} else {
					realDelay = delays[i] + ports.get(i).output.getQueueDelay();
				}
				
				if (realDelay < fastestTime || fastestPort == -1 ) {
					fastestPort = i;
					fastestTime = realDelay;
				}
			}
			
			if (fastestPort >= 0) {
				ports.get(fastestPort).output.push(packet);
			} else {
				packet.markAsNoRouteToHost();
			}
		}
	}
	
	/**
	 * Palauttaa reitittimen merkkiesityksen muodossa:
	 * 
	 *    ports[0].output
	 *    ...
	 *    ports[n].output
	 */
	public String toString() {
		StringBuffer s = new StringBuffer();
		//s.append("Router "+getID()+"\n");
		for (int i = 0; i < ports.size(); i++) {
			s.append("   ");
			s.append(ports.get(i));
			s.append("\n");
		}
		return s.toString();
	}
	
//	VANHENTUNUTTA KOODIA:
//
//	public static void main(String[] args){
//		
//		final int NODES = 5;
//		final int TICKS = 50;
//		final int START_NODE = 3;
//		final int END_NODE = 4;
//		
//		// luo lineaarinen verkko
//		Router[] r = new Router[NODES];
//		r[0] = new Router(0);
//		for (int i = 1; i < r.length; i++) {
//			r[i] = new Router(i);
//			r[i].join(r[i-1]);
//		}
//		
//		// luo rengas
//		r[0].join(r[r.length-1]);
//		
//		// pivit reititystaulut
//		for (int i = 0; i < r.length; i++) {
//			r[i].sendRouteMapPacket();
//		}
//		
//		// laita paketti etenemn ja seuraa
//		for (int tick = 0; tick < TICKS; tick++) {
//			Packet packet = new Packet(START_NODE, END_NODE, Packet.FASTEST_ROUTE, 64);
//			r[START_NODE].routePacket(packet);
//			
//			for (int i = 0; i < r.length; i++) {
//				System.out.println(r[i]);
//			}
//			System.out.println("-------- Tick "+tick+" ---------\n");
//			Ticker.tick();
//		}
//	}
}
