// Tietorakenteiden harjoitusty (syksy 2004), Esko Luontola
// Tietoliikenneverkon reititysalgoritmien vertailu, "Sim Router"

/**
 * Yhdist reitittimi toisiinsa huolehtien raja-arvoista. Metodeita kytettv
 * tss jrjestyksess:
 * 1. addRouter (n kertaa)
 * 2. makeConnections (1 kerta)
 * 3. makeComplete (1 kerta)
 * jonka jlkeen tll oliolla ei en tee mitn.
 */
public class RouterJoiner {
	
	/**
	 * Thn menness luotujen yhteyksien lukumr.
	 */
	private int connections;
	
	/**
	 * Toivottu lukumr yhteyksi. Todellinen lukumr voi menn tmn ali
	 * tai yli, jos muutoin yhtenisen verkon luominen ei onnistu.
	 */
	private int maxConnections;
	
	/**
	 * Reitittimet prioriteettijonossa, eniten yhteksi omaava reititin
	 * ensimmisen. Kun reititin saavutaa maksimimrns yhteyksi, se on
	 * siirrettv complete-listaan.
	 */
	private RouterJoinHeap incomplete;
	
	/**
	 * Lista reitittimist, jotka ovat saavuttaneet maksimimrns yhteyksi.
	 */
	private RouterVector complete;
	
	/**
	 * Luo uuden RouterJoinerin, joka tekee annetun mrn yhteyksi
	 */
	public RouterJoiner(int maxConnections) {
		this.connections = 0;
		this.maxConnections = maxConnections;
		this.incomplete = new RouterJoinHeap();
		this.complete = new RouterVector();
	}
	
	/**
	 * Palautaa verkkoon luotujen yhteyksien lukumrn.
	 */
	public int getConnections() {
		return connections;
	}
	
	/**
	 * Lis reitittimen incomplete-jonoon ja liitt sen johonkin toiseen reitittimeen.
	 */
	public void addRouter(Router router) {
		if (incomplete.size() == 0) {
			incomplete.push(router);
		} else {
			int other = (int)(incomplete.size() * Math.random());
			if (incomplete.joinRouters(router, other)) {
				connections++;
			}
			incomplete.push(router);
			joinCleanup();
		}
	}
	
	/**
	 * Luo enintn maxConnections verran yhteyksi, ehdolla ett kaikki
	 * reitittimet eivt saavuta yhteyksien maksimimr. 
	 * Jos maxConnections >= 2, niin verkosta tulee automaattisesti yhteninen.
	 */
	public void makeConnections() {
		int tries = 0;
		
		while (incomplete.size() > 0
			&& connections < maxConnections
			&& tries < 50) {
			
			int one = (int)(incomplete.size() * Math.random());
			int two = (int)(incomplete.size() * Math.random());
			tries++;
			
			if (incomplete.joinRouters(one, two)) {
				connections++;
				tries = 0;
			} else {
				// etsi jokin toinen, johon saadaan yhdistetty
				int i = (two + 1) % incomplete.size();
				while (i != two) {
					if (incomplete.joinRouters(one, i)) {
						connections++;
						tries = 0;
						break;
					}
					i = (i + 1) % incomplete.size();
				}
			}
			joinCleanup();
		}
	}
	
	/**
	 * Tekee verkosta yhtenisen ja alustaa reititystaulut. Kutsutaan 
	 * makeConnections():n jlkeen. Reitittimien yhteysrajat saattavat ylitty.
	 * Palauttaa true, jos verkosta tuli yhteninen.
	 */
	public boolean makeComplete() {
		// tyhjenn incomplete-jono ja siirr kaikki complete-jonoon
		while (incomplete.size() > 0) {
			complete.add(incomplete.pop());
		}
		
		ProgressBar.getCurrent().setMinimum(0);
		ProgressBar.getCurrent().setMaximum(complete.size());
		ProgressBar.getCurrent().setValue(0);
		
		// tarkista verkon yhtenisyys
		if (complete.size() > 0) {
			Router first = complete.get(0);
			first.sendRouteMapPacket();
			Ticker.tickUntilStopped();
			boolean wasFixed = false;
			
			for (int i = 1; i < complete.size(); i++) {
				if (!complete.get(i).isRouteToHost(first.getID())) {
					/* HUOM:
					 * Jos Simulation.getMaxPorts() >= 2, niin thn lohkoon
					 * EI JOUDUTA KOSKAAN, koska makeConnections() tekee jo 
					 * yhtenisen verkon. Siksi tm verkon yhtenisyyden 
					 * korjausalgori on tohdittu jtt hieman tehottomaksi,
					 * koska getMaxPorts() on aina >= 2.
					 */
					
					// yhdist toiseen, joka on jo osa yhtenist verkkoa
					int other = (int)(i * Math.random());
					if (complete.get(i).join(complete.get(other), true)) {
						connections++;
						wasFixed = true;
						System.out.println("RouterJoiner.makeComplete(): network fixup OK "+i+"<->"+other);
					} else {
						System.out.println("RouterJoiner.makeComplete(): network fixup FAILED "+i+"<->"+other);
					}
					
					// nollaa vanhentuneet reititystaulut
					for (int j = 0; j <= i; j++) {
						complete.get(j).resetRouteTables();
					}
					
					// tarkista yhtenisyys
					first.sendRouteMapPacket();
					if (!complete.get(i).isRouteToHost(first.getID())) {
						// jotain on pahasti pieless, jos vielkn ei ole yhteninen
						System.out.println("ERROR1: RouterJoiner.makeComplete()");
						return false;
					}
				}
			}
			
			// tyhjenn muutosten jljilt puutteelliset reititystaulut
			if (wasFixed) {
				for (int i = 0; i < complete.size(); i++) {
					complete.get(i).resetRouteTables();
				}
			}
		}
		
		// alusta kaikki reititystaulut
		for (int i = 0; i < complete.size(); i++) {
			ProgressBar.getCurrent().setValue(i);
			complete.get(i).sendRouteMapPacket();
			Ticker.tickUntilStopped();
			// DEBUG:
//			for (int j = 0; j < complete.size(); j++) {
//				System.out.println(complete.get(j).isRouteToHost(complete.get(i).getID()));
//			}
		}
		
		ProgressBar.getCurrent().setValue(complete.size());
		
		return true;
	}
	
	/**
	 * Poistaa incomplete-jonosta sellaiset reitittimet, jotka ovat saavuttaneet
	 * yhteksien maksimimrn.
	 */
	private void joinCleanup() {
		while (incomplete.size() > 0
			&& incomplete.getFirst().getConnections() >= Simulation.getMaxPorts()) {
			complete.add(incomplete.pop());
		}
	}
}
