// Tietorakenteiden harjoitusty (syksy 2004), Esko Luontola
// Tietoliikenneverkon reititysalgoritmien vertailu, "Sim Router"

/**
 * Simulaatiota hallinnoiva luokka.
 */
public class Simulation {
	
	/**
	 * Verkon reitittimet.
	 */
	private static RouterVector routers = new RouterVector();
	
	/**
	 * Verkon reitittimien vlisten yhteyksien lukumr.
	 */
	private static int connections = 0;
	
	/**
	 * Suurin sallittu mr yhteyksi yhdelle reitittimelle.
	 */
	private static int maxPorts = 5;
	
	/**
	 * Pakettipuskureiden koko.
	 */
	private static int bufferSize = 10;
	
	/**
	 * Suurin sallittu tiedonsiirron kesto kahden reitittimen vlill.
	 */
	private static int maxTransferDelay = 5;
	
	/**
	 * Alustaa simulaation tekemll halutunlaisen verkon. Reitittimien vlisi
	 * yhteyksi pyritn tekemn annettu mr ehdolla, ett raja-arvot eivt
	 * ylity ja verkosta tulee yhteninen.
	 *
	 * @param	routerCount  reitittimien lukumr
	 * @param	connectionCount  reitittimien vlisten yhteyksien lukumr (arvio)
	 */
	public static synchronized void prepareSimulation(int routerCount, int connectionCount) {
		routers = new RouterVector();
		connections = 0;
		RouterJoiner joiner = new RouterJoiner(connectionCount);
		
		// luo reitittimet ja yhteydet
		for (int i = 0; i < routerCount; i++) {
			Router router = new Router(i);
			routers.add(router);
			joiner.addRouter(router);
		}
		
		joiner.makeConnections();
		joiner.makeComplete();
		connections = joiner.getConnections();
		
		// alusta ajastin
		Ticker.reset();
		
//		System.out.println("Created a network with "+getRouters()+" routers and "
//		                   +getConnections()+" connections.");
	}
	
	/**
	 * Ajaa simulaation ja palauttaa kertyt tilastot.
	 *
	 * @param	protocoll  pakettien reitityksess kytettv protokolla
	 * @param	ttl  pakettien Time To Live. Jos 0, niin rajaton.
	 * @param	packets  kuinka monta pakettia lhetetn packetsInterval aikayksikn vlein
	 * @param	packetsInterval  kuinka monen aikayksikn vlein lhetetn packets pakettia. Jos <1, niin kytetn 1.
	 * @param	preTicks  ennen tilastojen keruuta kytettv aika
	 * @param	ticks  yhden tilasto-otteen keruuseen kytettv aika
	 * @param	repeats  kuinka monta tilasto-otetta tehdn perkkin
	 */
	public static synchronized Statistics[] runSimulation(int protocoll,
	                                                      int ttl,
	                                                      int packets,
	                                                      int packetsInterval,
	                                                      int preTicks,
	                                                      int ticks,
	                                                      int repeats) {
		// alustukset
		flushPackets();
		runSimulation(protocoll, ttl, packets, packetsInterval, preTicks);
		
		// simuloi ja ker tilastot
		Statistics[] stats = new Statistics[repeats];
		for (int i = 0; i < stats.length; i++) {
			stats[i] = runSimulation(protocoll, ttl, packets, packetsInterval, ticks);
		}
		return stats;
	}
	
	/**
	 * Ajaa simulaation ja palauttaa kertyt tilastot.
	 *
	 * @param	protocoll  pakettien reitityksess kytettv protokolla
	 * @param	ttl  pakettien Time To Live. Jos 0, niin rajaton.
	 * @param	packets  kuinka monta pakettia lhetetn packetsInterval aikayksikn vlein
	 * @param	packetsInterval  kuinka monen aikayksikn vlein lhetetn packets pakettia. Jos <1, niin kytetn 1.
	 * @param	ticks  simulaation pituus aikayksikkin
	 */
	public static synchronized Statistics runSimulation(int protocoll,
	                                                    int ttl,
	                                                    int packets,
	                                                    int packetsInterval,
	                                                    int ticks) {
		if (ttl < 0) {
			ttl = 0;
		}
		if (packetsInterval < 1) {
			packetsInterval = 1;
		}
		
		// alusta tilastot
		Statistics stats = Statistics.beginNew();
		
		// alusta ajastimet
		int endTime = Ticker.getTime() + ticks;
		int nextPackets = 0;
		int time;
		
		ProgressBar.getCurrent().setMinimum(Ticker.getTime());
		ProgressBar.getCurrent().setMaximum(endTime);
		
		while ((time = Ticker.getTime()) < endTime) {
			ProgressBar.getCurrent().setValue(time);
			
			// pakettien lhetys
			if (time >= nextPackets) {
				for (int i = 0; i < packets; i++) {
					int one = (int)(routers.size() * Math.random());
					int two = (one + (int)((routers.size()-1) * Math.random() + 1)) % routers.size();
					
					Packet packet = new Packet(one, two, protocoll, ttl);
					Statistics.registerPacket(packet);
					routers.get(one).routePacket(packet);
				}
				nextPackets += packetsInterval;
			}
			
			// kuluta aikaa
			if (SimRouter.DEBUG) {
				for (int i = 0; i < packetsInterval; i++) {
					Ticker.tick();
					printSimulation();
				}
			} else {
				Ticker.tick(packetsInterval);
			}
		}
		
		ProgressBar.getCurrent().setValue(endTime);
		
		Statistics.endCurrent();
		return stats;
	}
	
	/**
	 * Tyhjent simulaation kaikista liikkeell olevista paketeista ja nollaa ajastimen.
	 */
	public static void flushPackets() {
		Ticker.reset();
		for (int i = 0; i < routers.size(); i++) {
			routers.get(i).flushPackets();
		}
	}
	
	/**
	 * Simulaatiossa olevien reitittimien lukumr.
	 */
	public static int getRouters() {
		return routers.size();
	}
	
	/**
	 * Simulaatiossa olevien reitittimien vlisten yhteyksien lukumr.
	 */
	public static int getConnections() {
		return connections;
	}
	
	/**
	 * Asettaa suurimman sallitun mrn yhteyksi yhdelle reitittimelle. Jos
	 * uusi arvo on pienempi kuin 2, ei tehd mitn.
	 */
	public static void setMaxPorts(int value) {
		if (value >= 2) {
			maxPorts = value;
		}
	}
	
	/**
	 * Palauttaa suurimman sallitun mrn yhteyksi yhdelle reitittimelle.
	 */
	public static int getMaxPorts() {
		return maxPorts;
	}
	
	/**
	 * Asettaa pakettipuskureiden koon. Jos uusi arvo on pienempi kuin 1, 
	 * ei tehd mitn.
	 */
	public static void setBufferSize(int value) {
		if (value >= 1) {
			bufferSize = value;
		}
	}
	
	/**
	 * Palauttaa pakettipuskureiden koon.
	 */
	public static int getBufferSize() {
		return bufferSize;
	}
	
	/**
	 * Asettaa ylrajan kahden reitittimen vlisen tiedonsiirron kestolle. Jos
	 * uusi arvo on pienempi kuin 1, ei tehd mitn.
	 */
	public static void setMaxTransferDelay(int value) {
		if (value >= 1) {
			maxTransferDelay = value;
		}
	}
	
	/**
	 * Palauttaa ylrajan kahden reitittimen vlisen tiedonsiirron kestolle.
	 */
	public static int getMaxTransferDelay() {
		return maxTransferDelay;
	}
	
	/**
	 * Satunnainen luku vlilt 1 ja getMaxTransferDelay().
	 */
	public static int getRandomTransferDelay() {
		return (int)(getMaxTransferDelay() * Math.random()) + 1;
	}
	
	/**
	 * Tulostaa simulaation tmnhetkisen tilan.
	 */
	public static void printSimulation() {
		System.out.println("-------- Tick "+Ticker.getTime()+" ---------\n");
		for (int i = 0; i < routers.size(); i++) {
			System.out.println(routers.get(i));
		}
		System.out.println();
	}
	
	public static void main(String[] args){
		
		Simulation.setMaxPorts(3);
		Simulation.setBufferSize(10);
		Simulation.setMaxTransferDelay(5);
		
		// int routerCount, int connectionCount
		Simulation.prepareSimulation(500, 700);
		
		int protocoll;
		//protocoll = Packet.FLOOD;
		//protocoll = Packet.HOT_POTATO;
		protocoll = Packet.SHORTEST_ROUTE;
		//protocoll = Packet.FASTEST_ROUTE;
		
		// int protocoll, int ttl, int packets, int packetsInterval, int preTicks, int ticks, int repeats
		Statistics[] stats = Simulation.runSimulation(protocoll, 64, 500, 5, 1000, 1000, 5);
		
		for (int i = 0; i < stats.length; i++) {
			System.out.println("--- Statistics "+(i+1)+" ---");
			System.out.println(stats[i]);
		}
		
		// int protocoll, int ttl, int packets, int packetsInterval, int ticks
		Statistics s;
		s = Simulation.runSimulation(protocoll, 64, 500, 5, 1000);
		System.out.println(s);
		s = Simulation.runSimulation(protocoll, 64, 500, 5, 1000);
		System.out.println(s);
		s = Simulation.runSimulation(protocoll, 64, 500, 5, 1000);
		System.out.println(s);
	}
}
